import sys
import numpy as np
from .vireo_base import normalize, greed_match, donor_select
from .vireo_model_v02 import vireo_core


def vireo_flock(AD, DP, GT_prior=None, n_donor=None, n_extra_donor=2, 
               extra_donor_mode="distance", n_init=20, random_seed=None, 
               check_doublet=True, **kwargs):
    """
    A wrap function to run vireo twice, with the first step searching many 
    initialations.
    """
    ## random seed setting
    if random_seed is not None:
        np.random.seed(random_seed)

    if n_donor is None:
        if GT_prior is None:
            print("[vireo] Error: requiring n_donor or GT_prior.")
            sys.exit()
        else:
            n_donor = GT_prior.shape[2]

    ## warm initialization optionally with extra components
    _ID_prob = None
    if GT_prior is None or n_donor > GT_prior.shape[1]:
        n_donor_run1 = int(n_donor + n_extra_donor)
        print("[vireo] warm-up: %d random initializations for %d clusters..." 
              %(n_init, n_donor_run1))

        ID_prob_list = []
        for i in range(n_init):
            _ID_prob = np.random.rand(AD.shape[1], n_donor_run1)
            ID_prob_list.append(normalize(_ID_prob))

        result = []
        for _ID_prob in ID_prob_list:
            result.append(vireo_core(AD, DP, n_donor=n_donor_run1,
                GT_prior = None, ID_prob_init=_ID_prob, min_iter=5, max_iter=15, 
                verbose=False, check_doublet=False, **kwargs))

        LB_list = [x['LB_list'][-1] for x in result]
        res1 = result[np.argmax(LB_list)]

        _ID_prob = donor_select(res1['GT_prob'], res1['ID_prob'], n_donor, 
                                mode=extra_donor_mode)

        print("[vireo] warm-up: lower bound ranges [%.1f, %.1f, %.1f]" 
              %(min(LB_list), np.median(LB_list), max(LB_list)))
    else:
        _ID_prob = None

    ## pre-run: tune the genotype prior (if genotype prior is uncertain)
    GT_prior_use = GT_prior
    if GT_prior is not None and n_donor < GT_prior.shape[1]:
        print("[vireo] pre-RUN: finding %d from %d donors with GT ..." 
              %(n_donor, GT_prior.shape[2]))
        res1 = vireo_core(AD, DP, GT_prior = GT_prior, n_donor=None, 
                          ID_prob_init=_ID_prob, check_doublet=False, **kwargs)
        _donor_cnt = np.sum(res1['ID_prob'], axis=0)
        _donor_idx = np.argsort(_donor_cnt)[::-1]
        GT_prior_use = GT_prior[:, _donor_idx[:n_donor], :]

        print("\t".join(["donor%d" %x for x in _donor_idx]))
        print("\t".join(["%.0f" %_donor_cnt[x] for x in _donor_idx]))

    elif GT_prior is not None and n_donor > GT_prior.shape[1]:
        print("[vireo] pre-RUN: finding %d from %d donors without GT ..." 
              %(n_donor - GT_prior.shape[1], n_donor))
        res1 = vireo_core(AD, DP, GT_prior=None, n_donor=n_donor, 
                          ID_prob_init=_ID_prob, check_doublet=check_doublet, 
                          **kwargs)
        GT_prior_use = res1['GT_prob']
        idx = greed_match(GT_prior, GT_prior_use)
        GT_prior_use[:, idx, :] = GT_prior
        _idx_order = np.append(idx, np.delete(np.arange(n_donor), idx))
        GT_prior_use = GT_prior_use[:, _idx_order, :]

        _donor_cnt = np.sum(res1['ID_prob'], axis=0) 
        idx_ordered = np.append(idx, np.delete(np.arange(n_donor), idx))
        print("\t".join(["donor%d" %x for x in idx_ordered]))
        print("\t".join(["%.0f" %_donor_cnt[x] for x in idx_ordered]))

    ## main run
    print("[vireo] main RUN with warm initials and tuned GT ...")
    res1 = vireo_core(AD, DP, GT_prior=GT_prior_use, n_donor=n_donor, 
                      ID_prob_init=_ID_prob, check_doublet=check_doublet, 
                      **kwargs)
    print("[vireo] main RUN: %d iterations; lower bound %.1f" 
          %(len(res1['LB_list']), res1['LB_list'][-1]))

    ## print the beta parameters
    print("[vireo] beta parameters for binomial rate:")
    np.set_printoptions(formatter={'float': lambda x: format(x, '.2f')})
    print(res1['theta_shapes'])

    return res1



# base functions for plotting

import numpy as np

WeiZhu_colors = np.array(['#4796d7', '#f79e54', '#79a702', '#df5858', '#556cab', 
                          '#de7a1f', '#ffda5c', '#4b595c', '#6ab186', '#bddbcf', 
                          '#daad58', '#488a99', '#f79b78', '#ffba00'])

def heat_matrix(X, yticks=None, xticks=None, rotation=45, cmap='BuGn', 
                alpha=0.6, display_value=True, row_sort=False, 
                aspect='auto', **kwargs):
    """
    Plot heatmap of distance matrix
    """
    import matplotlib.pyplot as plt
    
    if row_sort:
        row_idx = np.argsort(np.dot(X, 2**np.arange(X.shape[1])))
        X = X[row_idx, :]

    im = plt.imshow(X, cmap=cmap, alpha=alpha, aspect=aspect, **kwargs)
    if xticks is not None:
        plt.xticks(range(len(xticks)), xticks, rotation=rotation)
        plt.xlim(-0.5, len(xticks) - 0.5)
    if yticks is not None:
        plt.yticks(range(len(yticks)), yticks)
        plt.ylim(-0.5, len(yticks) - 0.5)
    
    # Loop over data dimensions and create text annotations.
    if display_value:
        for i in range(X.shape[0]):
            for j in range(X.shape[1]):
                plt.text(j, i, "%.2f" %X[i, j],
                         ha="center", va="center", color="k")
    
    return im


def plot_GT(out_dir, cell_GPb, donor_names, 
            donor_GPb=None, donor_names_in=None):
    """
    Plot the genotype distance between samples
    """
    import matplotlib.pyplot as plt

    ## compare the GT probability of estimated samples
    diff_mat = np.zeros((cell_GPb.shape[1], cell_GPb.shape[1]))
    for i in range(cell_GPb.shape[1]):
        for j in range(cell_GPb.shape[1]):
            diff_mat[i,j] = np.mean(np.abs(cell_GPb[:, i, :] - 
                                           cell_GPb[:, j, :]))

    fig = plt.figure()
    heat_matrix(diff_mat, donor_names, donor_names)
    plt.title("Geno Prob Delta: %d SNPs" %(cell_GPb.shape[0]))
    plt.tight_layout()
    fig.savefig(out_dir + "/fig_GT_distance_estimated.pdf", dpi=300)

    ## compare in the estimated sample with input samples
    if donor_GPb is not None:
        diff_mat = np.zeros((cell_GPb.shape[1], donor_GPb.shape[1]))
        for i in range(cell_GPb.shape[1]):
            for j in range(donor_GPb.shape[1]):
                diff_mat[i,j] = np.mean(np.abs( cell_GPb[:, i, :] - 
                                               donor_GPb[:, j, :]))

        fig = plt.figure()
        heat_matrix(diff_mat, donor_names, donor_names_in)
        plt.title("Geno Prob Delta: %d SNPs" %(cell_GPb.shape[0]))
        plt.tight_layout()
        fig.savefig(out_dir + "/fig_GT_distance_input.pdf", dpi=300)


def minicode_plot(barcode_set, var_ids=None, sample_ids=None, 
                  cmap="Set3"):
    import matplotlib.pyplot as plt
    
    mat = np.zeros((len(barcode_set[0][1:]), len(barcode_set)))
    for i in range(mat.shape[0]):
        for j in range(mat.shape[1]):
            mat[i, j] = float(barcode_set[j][i + 1])
            
    im = plt.imshow(mat, cmap=cmap)
    for i in range(mat.shape[0]):
        for j in range(mat.shape[1]):
            plt.text(j, i, int(mat[i, j]), 
                     ha="center", va="center", color="k")
            
    if var_ids is None:
        var_ids = range(mat.shape[0])
    plt.yticks(range(len(var_ids)), var_ids)
    plt.ylim(-0.5, len(var_ids) - 0.5)
    
    if sample_ids is None:
        sample_ids = ["%s\nS%d" %(barcode_set[x], x)
                      for x in range(mat.shape[1])]
    else:
        sample_ids = ["%s\n%s" %(barcode_set[x], sample_ids[x])
                      for x in range(mat.shape[1])]
    plt.xticks(range(len(sample_ids)), sample_ids)
    plt.xlim(-0.5, len(sample_ids) - 0.5)

    return im

def anno_heat(X, anno, **kwargs):
    import seaborn as sns
    idx = np.argsort(np.dot(X, 2**np.arange(X.shape[1])) + 
                     anno * 2**X.shape[1])
    g = sns.clustermap(X[idx], cmap="GnBu", yticklabels=False,
                       col_cluster=False, row_cluster=False,
                       row_colors=WeiZhu_colors[anno][idx], **kwargs)
    
    for label in np.unique(anno):
        g.ax_col_dendrogram.bar(0, 0, color=WeiZhu_colors[label],
                                label=label, linewidth=0)
    g.ax_col_dendrogram.legend(loc="center", ncol=6, title="True clone")
    g.cax.set_position([.95, .2, .03, .45])
    return g

# def ppca_plot(AD, DP):
#     """
#     PPCA plot for each cell genotypes. This function is still underdevelopment
#     """
#     Z = DP.copy().astype(float)
#     idx = DP > 0
#     Z[idx] = AD[idx] / Z[idx]
#     Z[idx] = Z[idx] - 0.5

#     from sklearn.decomposition import TruncatedSVD
#     svd = TruncatedSVD(n_components=5, n_iter=7, random_state=42)
#     svd.fit(Z)

#     print("variance explained:", svd.explained_variance_ratio_)

#     import matplotlib.pyplot as plt
#     plt.scatter(svd.components_[0, :], svd.components_[1, :])
#     return svd.components_


# Utilility functions for processing vcf files
# Author: Yuanhua Huang
# Date: 24/06/2019

import os
import sys
import gzip
import subprocess
import numpy as np

def parse_sample_info(sample_dat, sparse=True, format_list=None):
    """
    Parse genotype information for each sample
    Note, it requires the format for each variants to 
    be the same.
    """
    if sample_dat == [] or sample_dat is None:
        return None

    # require the same format for all variants
    format_all = [x[0].split(":") for x in sample_dat]
    if format_list is None:
        format_list = format_all[0]

    RV = {}
    for _key in format_list:
        RV[_key] = []
    if sparse:
        ## sparse matrix requires all keys
        format_set_all = [set(x) for x in format_all]
        if format_set_all.count(set(format_list)) != len(format_all):
            print("Error: require the same format for all variants.")
            exit()

        RV['indices'] = []
        RV['indptr'] = [0]
        RV['shape'] = (len(sample_dat[0][1:]), len(sample_dat))
        missing_val = ":".join(["."] * len(format_list))
        
        cnt = 0
        for j in range(len(sample_dat)): #variant j
            _line = sample_dat[j]
            key_idx = [format_all[j].index(_key) for _key in format_list]
            for i in range(len(_line[1:])): #cell i
                if _line[i+1] == missing_val or _line[i+1] == ".":
                    continue
                _line_key = _line[i+1].split(":")
                for k in range(len(format_list)):
                    RV[format_list[k]].append(_line_key[key_idx[k]])

                cnt += 1
                RV['indices'].append(i)
            RV['indptr'].append(cnt)
    else:
        for j in range(len(sample_dat)): #variant j
            _line = sample_dat[j]
            _line_split = [x.split(":") for x in _line[1:]]
            for _key in format_list:
                if _key in format_all[j]:
                    k = format_all[j].index(_key)
                    _line_key = [x[k] for x in _line_split]
                    RV[_key].append(_line_key)
                else:
                    RV[_key].append(".")
    return RV


def load_VCF(vcf_file, biallelic_only=False, load_sample=True, sparse=True,
             format_list=None):
    """
    Load whole VCF file 
    -------------------
    Initially designed to load VCF from cellSNP output, requiring 
    1) all variants have the same format list;
    2) a line starting with "#CHROM", with sample ids.
    If these two requirements are satisfied, this function also supports general
    VCF files, e.g., genotype for multiple samples.

    Note, it may take a large memory, please filter the VCF with bcftools first.
    """
    if vcf_file[-3:] == ".gz" or vcf_file[-4:] == ".bgz":
        infile = gzip.open(vcf_file, "rb")
        is_gzip = True
    else:
        infile = open(vcf_file, "r")
        is_gzip = False
    
    FixedINFO = {}
    contig_lines = []
    comment_lines = []
    var_ids, obs_ids, obs_dat = [], [], []
    
    for line in infile:
        if is_gzip:
            line = line.decode('utf-8')
        if line.startswith("#"):
            if line.startswith("##contig="):
                contig_lines.append(line.rstrip())
            if line.startswith("#CHROM"):
                if load_sample:
                    obs_ids = line.rstrip().split("\t")[9:]
                key_ids = line[1:].rstrip().split("\t")[:8]
                for _key in key_ids:
                    FixedINFO[_key] = []
            else:
                comment_lines.append(line.rstrip())
        else:
            list_val = line.rstrip().split("\t") #[:5] #:8
            if biallelic_only:
                if len(list_val[3]) > 1 or len(list_val[4]) > 1:
                    continue
            if load_sample:
                obs_dat.append(list_val[8:])
            for i in range(len(key_ids)):
                FixedINFO[key_ids[i]].append(list_val[i])
            var_ids.append("_".join([list_val[x] for x in [0, 1, 3, 4]]))
    infile.close()

    RV = {}
    RV["variants"]  = var_ids
    RV["FixedINFO"] = FixedINFO
    RV["contigs"]   = contig_lines
    RV["comments"]  = comment_lines
    if load_sample:
        RV["samples"]   = obs_ids
        RV["GenoINFO"]  = parse_sample_info(obs_dat, sparse, format_list)
    return RV


def write_VCF_to_hdf5(VCF_dat, out_file):
    """
    Write vcf data into hdf5 file
    """
    import h5py
    f = h5py.File(out_file, 'w')
    f.create_dataset("contigs", data=np.string_(VCF_dat['contigs']), 
                     compression="gzip", compression_opts=9)
    f.create_dataset("samples", data=np.string_(VCF_dat['samples']), 
                     compression="gzip", compression_opts=9)
    f.create_dataset("variants", data=np.string_(VCF_dat['variants']), 
                     compression="gzip", compression_opts=9)
    f.create_dataset("comments", data=np.string_(VCF_dat['comments']), 
                     compression="gzip", compression_opts=9)
    
    ## variant fixed information
    fixed = f.create_group("FixedINFO")
    for _key in VCF_dat['FixedINFO']:
        fixed.create_dataset(_key, data=np.string_(VCF_dat['FixedINFO'][_key]), 
                             compression="gzip", compression_opts=9)
        
    ## genotype information for each sample
    geno = f.create_group("GenoINFO")
    for _key in VCF_dat['GenoINFO']:
        geno.create_dataset(_key, data=np.string_(VCF_dat['GenoINFO'][_key]), 
                             compression="gzip", compression_opts=9)
        
    f.close()


def read_sparse_GeneINFO(GenoINFO, keys=['AD', 'DP'], axes=[-1, -1]):
    M, N = np.array(GenoINFO['shape']).astype('int')
    indptr = np.array(GenoINFO['indptr']).astype('int')
    indices = np.array(GenoINFO['indices']).astype('int')
    
    from scipy.sparse import csr_matrix
    
    RV = {}
    for i in range(len(keys)):
        _dat = [x.split(",")[axes[i]] for x in GenoINFO[keys[i]]]
        _dat = [x if x != '.' else '0' for x in _dat]
        data = np.array(_dat).astype('float')
        RV[keys[i]] = csr_matrix((data, indices, indptr), shape=(N, M))
    return RV


def GenoINFO_maker(GT_prob, AD_reads, DP_reads):
    """
    Generate the Genotype information for estimated genotype probability at
    sample level.
    """
    GT_val = np.argmax(GT_prob, axis=2)
    GT_prob[GT_prob < 10**(-10)] = 10**(-10)
    PL_prob = np.round(-10 * np.log10(GT_prob)).astype(int).astype(str)
    AD_reads = np.round(AD_reads).astype(int).astype(str)
    DP_reads = np.round(DP_reads).astype(int).astype(str)

    GT, PL, AD, DP = [], [], [], []
    for i in range(GT_prob.shape[0]):
        GT.append([['0/0', '1/0', '1/1'][x] for x in GT_val[i, :]])
        PL.append([",".join(list(x)) for x in PL_prob[i, :, :]])
        AD.append(list(AD_reads[i, :]))
        DP.append(list(DP_reads[i, :]))
    
    RV = {}
    RV['GT'] = GT
    RV['AD'] = AD
    RV['DP'] = DP
    RV['PL'] = PL
    return RV


def write_VCF(out_file, VCF_dat, GenoTags=['GT', 'AD', 'DP', 'PL']):
    if out_file.endswith(".gz"):
        out_file_use = out_file.split(".gz")[0]
    else:
        out_file_use = out_file
        
    fid_out = open(out_file_use, "w")
    for line in VCF_dat['comments']:
        fid_out.writelines(line + "\n")
    
    VCF_COLUMN = ["CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", 
                  "INFO", "FORMAT"]
    fid_out.writelines("#" + "\t".join(VCF_COLUMN + VCF_dat['samples']) + "\n")
    
    for i in range(len(VCF_dat['variants'])):
        line = [VCF_dat['FixedINFO'][x][i] for x in VCF_COLUMN[:8]]
        line.append(":".join(GenoTags))

        # for d in range(len(VCF_dat['GenoINFO'][GenoTags[0]][0])):
            # _line_tag = [VCF_dat['GenoINFO'][x][i][d] for x in GenoTags]
        for s in range(len(VCF_dat['samples'])):
            _line_tag = [VCF_dat['GenoINFO'][_tag][i][s] for _tag in GenoTags]
            line.append(":".join(_line_tag))
        fid_out.writelines("\t".join(line) + "\n")
    fid_out.close()

    import shutil
    if shutil.which("bgzip") is not None:
        bashCommand = "bgzip -f %s" %(out_file_use)
    else:
        bashCommand = "gzip -f %s" %(out_file_use)
    pro = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    pro.communicate()[0]

    
def parse_donor_GPb(GT_dat, tag='GT', min_prob=0.0):
    """
    Parse the donor genotype probability
    tag: GT, GP, or PL
    """
    def parse_GT_code(code, tag, min_prob=0):
        if code == "." or code == "./." or code == ".|.":
            return np.array([1/3, 1/3, 1/3])
        if tag == 'GT':
            _prob = np.array([0, 0, 0], float)
            _prob[int(float(code[0]) + float(code[-1]))] = 1
        elif tag == 'GP':
            _prob = np.array(code.split(','), float)
        elif tag == 'PL':
            _Phred = np.array(code.split(','), float)
            _prob = 10**(-0.1 * (_Phred - min(_Phred)) - 0.025) # 0?
        else:
            _prob = None
            
        _prob += min_prob
        return _prob / np.sum(_prob)

    if ['GT', 'GP', 'PL'].count(tag) == 0:
        print("[parse_donor_GPb] Error: no support tag: %s" %tag)
        return None

    GT_prob = np.zeros((len(GT_dat), len(GT_dat[0]), 3))
    for i in range(GT_prob.shape[0]):
        for j in range(GT_prob.shape[1]):
            GT_prob[i, j, :] = parse_GT_code(GT_dat[i][j], tag,
                                             min_prob)
    return GT_prob



import itertools
import numpy as np
from scipy.stats import entropy
from scipy.special import logsumexp, digamma, betaln
from .vireo_base import normalize, loglik_amplify, beta_entropy, get_binom_coeff

__docformat__ = "restructuredtext en"

__all__ = ['Vireo', 'add_doublet_theta', 'add_doublet_GT']


class Vireo():
    """Viroe model: Variational Inference for reconstruction of ensemble origin

    The prior can be set via set_prior() before fitting the model.

    Key properties
    --------------
    beta_mu: numpy array (1, n_GT) or (n_var, n_GT)
        Beta mean parameter of theta's posterior
    beta_sum: numpy array (1, n_GT) or (n_var, n_GT), same as beta_mu
        Beta concetration parameter of theta's posterior
    ID_prob: numpy array (n_cell, n_donor)
        Posterior cell assignment probability to each donor
    GT_prob: numpy array (n_var, n_donor, n_GT)
        Posterior genotype probability per variant per donor
    """
    def __init__(self, n_cell, n_var, n_donor, n_GT=3, learn_GT=True,  
        learn_theta=True, ASE_mode=False, fix_beta_sum=False, 
        beta_mu_init=None, beta_sum_init=None, ID_prob_init=None, 
        GT_prob_init=None):
        """Initialise Vireo model

        Note, multiple initializations are highly recomended to avoid local 
        optima.
        
        Parameters
        ----------
        n_cell : int. 
            Number of cells
        n_var : int. 
            Number of variants
        n_donor : int. 
            Number of donors
        n_GT : int. 
            Number of genotype categories
        learn_GT: bool. 
            Whether updating `GT_prob`; otherwise using the initial
        ASE_mode: bool. 
            Whether setting allelic ratio `theta` to be variant specific
        fix_beta_sum: bool. 
            Whether fixing the concetration parameter of theta's posterior
        beta_mu_init: numpy array (1, n_GT) or (n_var, n_GT)
            Initial value of beta_mu, the mean parameter of theta
        beta_sum_init: numpy array (1, n_GT) or (n_var, n_GT), same as beta_mu
            Initial value of beta_sum, the concetration parameter of theta
        ID_prob_init: numpy array (n_cell, n_donor)
            Initial value of ID_prob, cell assignment probability to each donor
        GT_prob_init: numpy array (n_var, n_donor, n_GT)
            Initial value of GT_prob, genotype probability per variant and donor
        """
        self.n_GT = n_GT
        self.n_var = n_var
        self.n_cell = n_cell
        self.n_donor = n_donor
        self.learn_GT = learn_GT
        self.ASE_mode = ASE_mode
        self.learn_theta = learn_theta
        self.fix_beta_sum = fix_beta_sum
        
        theta_len = n_var if ASE_mode else 1
        
        # initial key parameters
        if beta_mu_init is not None:
            self.beta_mu = beta_mu_init
        else:
            self.beta_mu = (np.ones((theta_len, n_GT)) * 
                np.linspace(0.01, 0.99, n_GT).reshape(1, -1))

        if beta_sum_init is not None:
            self.beta_sum = beta_sum_init
        else:
            self.beta_sum = np.ones((theta_len, n_GT)) * 50

        if ID_prob_init is not None:
            self.ID_prob = normalize(ID_prob_init, axis=1)
        else:
            self.ID_prob = normalize(np.random.rand(n_cell, n_donor))

        if GT_prob_init is not None:
            self.GT_prob = normalize(GT_prob_init)
        else:        
            self.GT_prob = normalize(np.random.rand(n_var, n_donor, n_GT))

        self.ELBO_ = np.zeros((0))
        self.set_prior()
    
    def set_prior(self, GT_prior=None, ID_prior=None, beta_mu_prior=None, 
        beta_sum_prior=None, min_GP=0.00001):
        """Set prior for key variables: theta, GT_prob and ID_prob.
        The priors are in the same shape as its according variables.

        min_GP: float. Minimun genotype probability in GT_prior.
        """
        if beta_mu_prior is None:
            beta_mu_prior = np.expand_dims(
                np.linspace(0.01, 0.99, self.beta_mu.shape[1]), axis=0)
        if beta_sum_prior is None:
            beta_sum_prior = np.ones(beta_mu_prior.shape) * 50.0
        self.theta_s1_prior = beta_mu_prior * beta_sum_prior
        self.theta_s2_prior = (1 - beta_mu_prior) * beta_sum_prior

        if ID_prior is not None:
            if len(ID_prior.shape) == 1:
                ID_prior = np.expand_dims(ID_prior, axis=0)
            self.ID_prior = ID_prior
        else:
            self.ID_prior = normalize(np.ones(self.ID_prob.shape))

        if GT_prior is not None:
            if len(GT_prior.shape) == 2:
                GT_prior = np.expand_dims(GT_prior, axis=0)
            GT_prior[GT_prior < min_GP] = min_GP
            GT_prior[GT_prior > 1 - min_GP] = 1 - min_GP
            GT_prior = normalize(GT_prior)
            self.GT_prior = GT_prior
        else:        
            self.GT_prior = normalize(np.ones(self.GT_prob.shape))

    @property
    def theta_s1(self):
        """Beta concetration1 parameter for theta posterior"""
        return self.beta_mu * self.beta_sum

    @property
    def theta_s2(self):
        """Beta concetration2 parameter for theta posterior"""
        return (1 - self.beta_mu) * self.beta_sum

    @property 
    def digamma1_(self):
        """Digamma of Beta concetration1 parameter"""
        return np.expand_dims(digamma(self.theta_s1), 1)

    @property 
    def digamma2_(self):
        """Digamma of Beta concetration2 parameter"""
        return np.expand_dims(digamma(self.theta_s2), 1)

    @property 
    def digammas_(self):
        """Digamma of Beta concetration summary parameter"""
        return np.expand_dims(digamma(self.theta_s1 + self.theta_s2), 1)


    def update_theta_size(self, AD, DP):
        """Coordinate ascent for updating theta posterior parameters
        """
        BD = DP - AD
        S1_gt = AD * self.ID_prob  #(n_var, n_donor)
        S2_gt = BD * self.ID_prob  #(n_var, n_donor)
        
        _theta_s1 = np.zeros(self.beta_mu.shape)
        _theta_s2 = np.zeros(self.beta_mu.shape)
        _theta_s1 += self.theta_s1_prior.copy()
        _theta_s2 += self.theta_s2_prior.copy()
        for ig in range(self.n_GT):
            _axis = 1 if self.ASE_mode else None
            _theta_s1[:, ig:(ig+1)] += np.sum(
                S1_gt * self.GT_prob[:, :, ig], axis=_axis, keepdims=True)
            _theta_s2[:, ig:(ig+1)] += np.sum(
                S2_gt * self.GT_prob[:, :, ig], axis=_axis, keepdims=True)
        
        self.beta_mu = _theta_s1 / (_theta_s1 + _theta_s2)
        if self.fix_beta_sum == False:
            self.beta_sum = _theta_s1 + _theta_s2

    def update_ID_prob(self, AD, DP):
        """Coordinate ascent for updating assignment probability
        """
        BD = DP - AD
        logLik_ID = np.zeros((AD.shape[1], self.n_donor))
        for ig in range(self.n_GT):
            S1 = AD.transpose() * (self.GT_prob[:, :, ig] * self.digamma1_[:, :, ig])
            S2 = BD.transpose() * (self.GT_prob[:, :, ig] * self.digamma2_[:, :, ig])
            SS = DP.transpose() * (self.GT_prob[:, :, ig] * self.digammas_[:, :, ig])
            logLik_ID += (S1 + S2 - SS)
        
        self.ID_prob = normalize(np.exp(loglik_amplify(
            logLik_ID + np.log(self.ID_prior))))
        
        return logLik_ID
                

    def update_GT_prob(self, AD, DP):
        """Coordinate ascent for updating genotype probability
        """
        S1_gt = AD * self.ID_prob
        SS_gt = DP * self.ID_prob
        S2_gt = SS_gt - S1_gt
        
        logLik_GT = np.zeros(self.GT_prior.shape)
        for ig in range(self.n_GT):
            logLik_GT[:, :, ig] = (
                S1_gt * self.digamma1_[:, :, ig] + 
                S2_gt * self.digamma2_[:, :, ig] - 
                SS_gt * self.digammas_[:, :, ig])

        self.GT_prob = normalize(np.exp(loglik_amplify(
            logLik_GT + np.log(self.GT_prior))))
        
        
    def get_ELBO(self, logLik_ID, AD=None, DP=None):
        """Calculating variational evidence lower bound with current parameters

        logLik_ID: numpy array (n_cell, n_donor), the output from update_ID_prob
        """
        if logLik_ID is None:
            BD = DP - AD
            logLik_ID = np.zeros((AD.shape[1], self.n_donor))
            for ig in range(self.n_GT):
                S1 = AD.transpose() * (self.GT_prob[:, :, ig] * self.digamma1_[:, :, ig])
                S2 = BD.transpose() * (self.GT_prob[:, :, ig] * self.digamma2_[:, :, ig])
                SS = DP.transpose() * (self.GT_prob[:, :, ig] * self.digammas_[:, :, ig])
                logLik_ID += (S1 + S2 - SS)

        LB_p = np.sum(logLik_ID * self.ID_prob)
        KL_ID = -np.sum(entropy(self.ID_prob, self.ID_prior, axis=-1))
        KL_GT = -np.sum(entropy(self.GT_prob, self.GT_prior, axis=-1))
        KL_theta = -beta_entropy(
            np.append(
                np.expand_dims(self.theta_s1, 1), 
                np.expand_dims(self.theta_s2, 1), axis = 1),
            np.append(
                np.expand_dims(self.theta_s1_prior, 1), 
                np.expand_dims(self.theta_s2_prior, 1), axis = 1))
        
        # print(LB_p, KL_ID, KL_GT, KL_theta)
        return LB_p - KL_ID - KL_GT - KL_theta


    def fit(self, AD, DP, max_iter=200, min_iter=5, epsilon_conv=1e-2,
        delay_fit_theta=0, verbose=False):
        """Fit Vireo model with coordinate ascent

        Parameters
        ----------
        AD : scipy.sparse.csc_matrix (n_var, n_cell)
            Sparse count matrix for alternative allele
        DP : scipy.sparse.csc_matrix (n_var, n_cell)
            Sparse count matrix for depths, alternative + refeerence alleles
        max_iter : int
            Maximum number of iterations
        min_iter :
            Minimum number of iterations
        epsilon_conv : float
            Threshold for detecting convergence
        delay_fit_theta : int
            Number of steps to delay updating theta. This can be very useful 
            for common genetics when there is good prior on allelic ratio.
        verbose : bool
            Whether print out log info
        """
        _binom_coeff = np.sum(get_binom_coeff(AD, DP))
        ELBO = np.zeros(max_iter)
        for it in range(max_iter):
            if self.learn_theta and it >= delay_fit_theta:
                self.update_theta_size(AD, DP)
            if self.learn_GT:
                self.update_GT_prob(AD, DP)

            _logLik_ID = self.update_ID_prob(AD, DP)
            ELBO[it] = self.get_ELBO(_logLik_ID) + _binom_coeff

            if it > min_iter:
                if ELBO[it] < ELBO[it - 1]:
                    if verbose:
                        print("Warning: Lower bound decreases!\n")
                elif it == max_iter - 1:
                    if verbose:
                        print("Warning: VB did not converge!\n")
                elif ELBO[it] - ELBO[it - 1] < epsilon_conv:
                    break
        self.ELBO_ = np.append(self.ELBO_, ELBO[:it])


    def predict_doublet(self, AD, DP, update_GT=True, update_ID=True, 
        doublet_rate_prior=None):
        """Predict doublet with fitted Vireo model

        Parameters
        ----------
        AD : scipy.sparse.csc_matrix (n_var, n_cell)
            Sparse count matrix for alternative allele
        DP : scipy.sparse.csc_matrix (n_var, n_cell)
            Sparse count matrix for depths, alternative + refeerence alleles
        update_GT : bool
            Whether updating GT_prob after removing doublet_prob
        update_GT : bool
            Whether updating ID_prob by removing doublet_prob
        doublet_rate_prior : float
            Prior value of doublet rate

        Returns
        -------
        A tuple of two numpy arrays (doublet_prob, ID_prob)

        doublet_prob : numpy array (n_cell, n_donor * (n_donor - 1) / 2)
            Assignment probability of a cells to any doublet (donor pair)
        ID_prob : numpy array (n_cell, n_donor)
            updated ID_prob by removing doublet_prob
        """
        GT_both = add_doublet_GT(self.GT_prob)
        beta_mu_both, beta_sum_both = add_doublet_theta(self.beta_mu, 
                                                        self.beta_sum)

        n_doublet_pair = GT_both.shape[1] - self.GT_prob.shape[1]
        if doublet_rate_prior is None:
            doublet_rate_prior = min(0.5, AD.shape[1] / 100000)
            
        ID_prior_both = np.append(
            self.ID_prior * (1 - doublet_rate_prior), 
            np.ones((self.n_cell, n_doublet_pair)) / n_doublet_pair * 
            doublet_rate_prior, axis=1)

        # Calculate assignment probability (same as update_ID_prob())
        BD = DP - AD
        logLik_ID = np.zeros((AD.shape[1], GT_both.shape[1]))
        _digamma1 = np.expand_dims(digamma(beta_sum_both * beta_mu_both), 1)
        _digamma2 = np.expand_dims(digamma(beta_sum_both * (1 - beta_mu_both)), 1)
        _digammas = np.expand_dims(digamma(beta_sum_both), 1)
        for ig in range(GT_both.shape[2]):
            S1 = AD.transpose() * (GT_both[:, :, ig] * _digamma1[:, :, ig])
            S2 = BD.transpose() * (GT_both[:, :, ig] * _digamma2[:, :, ig])
            SS = DP.transpose() * (GT_both[:, :, ig] * _digammas[:, :, ig])
            logLik_ID += (S1 + S2 - SS)

        ID_prob_both = normalize(np.exp(loglik_amplify(
            logLik_ID + np.log(ID_prior_both))))

        if update_ID:
            self.ID_prob = ID_prob_both[:, :self.n_donor]

        if update_GT:
            if update_ID:
                self.update_GT_prob(AD, DP)
            else:
                print("For update_GT, please turn on update_ID.")
        
        return ID_prob_both[:, self.n_donor:], ID_prob_both[:, :self.n_donor]


    def predit_ambient(self):
        """Predict fraction of ambient RNA contaimination.
        
        Not implemented yet.
        """
        print("Not implemented yet.")


def add_doublet_theta(beta_mu, beta_sum):
    """
    calculate theta for doublet genotype: GT=0&1, GT=0&2, and GT=1&2 by
    averaging thire beta paramters
    
    Example
    -------
    add_doublet_theta(np.array([[0.01, 0.5, 0.99]]), np.array([[30, 6, 30]]))
    """
    # TODO: support reduced GT for relatives
    combn_iter = itertools.combinations(range(beta_mu.shape[1]), 2)
    db_idx = np.array([x for x in combn_iter])

    beta_mu_db = (beta_mu[:, db_idx[:, 0]] + beta_mu[:, db_idx[:, 1]]) / 2.0
    beta_sum_db = np.sqrt(beta_sum[:, db_idx[:, 0]] * beta_sum[:, db_idx[:, 1]])

    return (np.append(beta_mu, beta_mu_db, axis=-1), 
            np.append(beta_sum, beta_sum_db, axis=-1))


def add_doublet_GT(GT_prob):
    """
    Add doublet genotype by summarizing their probability:
    New GT has five categories: 0, 1, 2, 1.5, 2.5
    TODO: New GT has six categories: 0, 1, 2, 0_1, 0_2, 1_2
    """
    combn_iter = itertools.combinations(range(GT_prob.shape[2]), 2)
    gt_idx = np.array([x for x in combn_iter]) # GT combination
    g_idx1 = gt_idx[:, 0]
    g_idx2 = gt_idx[:, 1]

    combn_iter = itertools.combinations(range(GT_prob.shape[1]), 2)
    sp_idx = np.array([x for x in combn_iter]) # sample combination
    s_idx1 = sp_idx[:, 0]
    s_idx2 = sp_idx[:, 1]
    
    ## GT_prob has three genotypes: 0, 1, 2;
    n_gt = GT_prob.shape[2]
    GT_prob2 = np.zeros((GT_prob.shape[0], sp_idx.shape[0],
                         n_gt + gt_idx.shape[0]))

    GT_prob2[:, :, :n_gt] = (GT_prob[:, s_idx1, :] * 
                             GT_prob[:, s_idx2, :])
    GT_prob2[:, :, n_gt:] = (GT_prob[:, s_idx1, :][:, :, g_idx1] * 
                             GT_prob[:, s_idx2, :][:, :, g_idx2] +
                             GT_prob[:, s_idx1, :][:, :, g_idx2] * 
                             GT_prob[:, s_idx2, :][:, :, g_idx1])
    
    GT_prob2 = normalize(GT_prob2, axis=2)
    GT_prob1 = np.append(GT_prob, 
        np.zeros((GT_prob.shape[0], GT_prob.shape[1], gt_idx.shape[0])), axis=2)
    return np.append(GT_prob1, GT_prob2, axis=1)



# Wrap function for Vireo model
# Author: Yuanhua Huang
# Date: 22/03/2020

import sys
import numpy as np
from .vireo_base import optimal_match, donor_select
from .vireo_model import Vireo


def vireo_wrap(AD, DP, GT_prior=None, n_donor=None, learn_GT=True, n_init=20, 
    random_seed=None, check_doublet=True, max_iter_init=20, delay_fit_theta=3,
    n_extra_donor=0, extra_donor_mode="distance", **kwargs):
    """
    A wrap function to run vireo with multiple initializations
    """
    if learn_GT == False and n_extra_donor > 0:
        print("Searching from extra donors only works with learn_GT")
        n_extra_donor = 0
        
    # note learn_GT is false for mode 2 and 5 only (set before)
    if n_donor is None:
        if GT_prior is None:
            print("[vireo] Error: requiring n_donor or GT_prior.")
            sys.exit()
        else:
            n_donor = GT_prior.shape[1]

    if learn_GT is False and n_init > 1:
        print("GT is fixed, so use a single initialization")
        n_init = 1

    ## Setting random seed for initialization
    if random_seed is not None:
        np.random.seed(random_seed)

    GT_prior_use = None
    n_donor_use = int(n_donor + n_extra_donor)
    if GT_prior is not None and n_donor_use == GT_prior.shape[1]:
        GT_prior_use = GT_prior.copy()
    elif GT_prior is not None and n_donor_use < GT_prior.shape[1]:
        GT_prior_use = GT_prior.copy()
        n_donor_use = GT_prior.shape[1]

    _models_all = []
    for im in range(n_init):
        _modelCA = Vireo(n_var=AD.shape[0], n_cell=AD.shape[1], 
                         n_donor=n_donor_use, learn_GT=learn_GT, 
                         GT_prob_init=GT_prior_use, **kwargs)
        _modelCA.set_prior(GT_prior=GT_prior_use)
        _models_all.append(_modelCA)

    ## Fitting the models
    for im in range(n_init):
        # _models_all[im].fit(AD, DP, min_iter=20, verbose=False)
        _models_all[im].fit(AD, DP, min_iter=5, max_iter=max_iter_init, 
            delay_fit_theta=delay_fit_theta, verbose=False)

    ## select the model with best initialization
    elbo_all = np.array([x.ELBO_[-1] for x in _models_all])
    _idx = np.argmax(elbo_all)
    modelCA = _models_all[_idx]
    if n_extra_donor == 0:
        modelCA.fit(AD, DP, min_iter=5, verbose=False)
    else:
        _ID_prob = donor_select(modelCA.GT_prob, modelCA.ID_prob, n_donor, 
                                mode=extra_donor_mode)
        modelCA = Vireo(n_var=AD.shape[0], n_cell=AD.shape[1], 
                        n_donor=n_donor, learn_GT=learn_GT, 
                        GT_prob_init=GT_prior_use, ID_prob_init=_ID_prob,
                        beta_mu_init=modelCA.beta_mu, 
                        beta_sum_init=modelCA.beta_sum, **kwargs)
        modelCA.set_prior(GT_prior=GT_prior_use)
        modelCA.fit(AD, DP, min_iter=5, delay_fit_theta=delay_fit_theta, 
            verbose=False)

    print("[vireo] lower bound ranges [%.1f, %.1f, %.1f]" 
          %(np.min(elbo_all), np.median(elbo_all), np.max(elbo_all)))

    ## Run Vireo again with updateing genotype
    if GT_prior is not None and n_donor < GT_prior.shape[1]:
        _donor_cnt = np.sum(modelCA.ID_prob, axis=0)
        _donor_idx = np.argsort(_donor_cnt)[::-1]
        GT_prior_use = GT_prior[:, _donor_idx[:n_donor], :]

        modelCA = Vireo(n_var=AD.shape[0], n_cell=AD.shape[1], 
                        n_donor=n_donor, learn_GT=False, 
                        GT_prob_init=GT_prior_use, **kwargs)
        modelCA.fit(AD, DP, min_iter=20, verbose=False)

    elif GT_prior is not None and n_donor > GT_prior.shape[1]:
        GT_prior_use = modelCA.GT_prob.copy()
        idx = optimal_match(GT_prior, GT_prior_use)[1]
        GT_prior_use[:, idx, :] = GT_prior
        _idx_order = np.append(idx, np.delete(np.arange(n_donor), idx))
        GT_prior_use = GT_prior_use[:, _idx_order, :]
        ID_prob_use = modelCA.ID_prob[:, _idx_order]

        modelCA = Vireo(n_var=AD.shape[0], n_cell=AD.shape[1], 
                        n_donor=n_donor, learn_GT=learn_GT,
                        ID_prob_init=ID_prob_use,
                        beta_mu_init=modelCA.beta_mu, 
                        beta_sum_init=modelCA.beta_sum,
                        GT_prob_init=GT_prior_use, **kwargs)
        modelCA.set_prior(GT_prior = GT_prior_use)
        modelCA.fit(AD, DP, min_iter=20, verbose=False)
    
    ## print the beta parameters
    print("[vireo] allelic rate mean and concentrations:")
    print(np.round(modelCA.beta_mu, 3))
    print(np.round(modelCA.beta_sum, 1))

    ## Summarise donor size
    print("[vireo] donor size before removing doublets:")
    _donor_cnt = np.sum(modelCA.ID_prob, axis=0)
    print("\t".join(["donor%d" %x for x in range(len(_donor_cnt))]))
    print("\t".join(["%.0f" %x for x in _donor_cnt]))
    
    
    ## Predict doublets
    if check_doublet:
        doublet_prob, ID_prob = modelCA.predict_doublet(AD, DP)
    else:
        ID_prob = modelCA.ID_prob
        doublet_prob = np.zeros((AD.shape[1], AD.shape[1] * (AD.shape[1] - 1) / 2))

    theta_shapes = np.append(modelCA.beta_mu * modelCA.beta_sum, 
                             (1 - modelCA.beta_mu) * modelCA.beta_sum, axis=0)
    RV = {}
    RV['ID_prob'] = ID_prob
    RV['GT_prob'] = modelCA.GT_prob
    RV['doublet_prob'] = doublet_prob
    RV['theta_shapes'] = theta_shapes
    RV['theta_mean'] = modelCA.beta_mu
    RV['theta_sum'] = modelCA.beta_sum
    RV['LB_list'] = elbo_all
    RV['LB_doublet'] = modelCA.ELBO_[-1]
    return RV




import subprocess
import numpy as np
from scipy.io import mmread
from itertools import permutations

from .vireo_base import match
from .vcf_utils import load_VCF, write_VCF, parse_donor_GPb
from .vcf_utils import read_sparse_GeneINFO, GenoINFO_maker


def match_donor_VCF(cell_dat, donor_vcf):
    """Match variants between cell VCF and donor VCF information
    """
    mm_idx = match(cell_dat['variants'], donor_vcf['variants'])
    mm_idx = mm_idx.astype(float)
    if np.sum(mm_idx == mm_idx) == 0 or np.sum(mm_idx >= 0) == 0:
        _cell_id = ["chr" + x for x in cell_dat['variants']]
        mm_idx = match(_cell_id, donor_vcf['variants'])
    if np.sum(mm_idx == mm_idx) == 0 or np.sum(mm_idx >= 0) == 0:
        _donor_id = ["chr" + x for x in donor_vcf['variants']]
        mm_idx = match(cell_dat['variants'], _donor_id)

    idx1 = np.where(mm_idx == mm_idx)[0] #remove None
    # TODO: check when chr is not compatible! given warning.
    if len(idx1) == 0:
        print("[vireo] warning: no variants matched to donor VCF, " + 
                "please check chr format!")
    else:
        print("[vireo] %d out %d variants matched to donor VCF" 
                %(len(idx1), len(cell_dat['variants'])))
    idx2 = mm_idx[idx1].astype(int)

    cell_dat['AD'] = cell_dat['AD'][idx1, :]
    cell_dat['DP'] = cell_dat['DP'][idx1, :]
    cell_dat["variants"]  = [cell_dat["variants"][x] for x in idx1]
    for _key in cell_dat["FixedINFO"].keys():
        cell_dat["FixedINFO"][_key] = [
            cell_dat["FixedINFO"][_key][x] for x in idx1]

    donor_vcf["variants"]  = [donor_vcf["variants"][x] for x in idx2]
    for _key in donor_vcf["FixedINFO"].keys():
        donor_vcf["FixedINFO"][_key] = [
            donor_vcf["FixedINFO"][_key][x] for x in idx2]
    for _key in donor_vcf["GenoINFO"].keys():
        donor_vcf["GenoINFO"][_key] = [
            donor_vcf["GenoINFO"][_key][x] for x in idx2]  

    return cell_dat, donor_vcf


def read_cellSNP(dir_name):
    """Read data from the cellSNP output directory

    Parameters
    ----------
    dir_name:
        directory full path name for cellSNP output
    
    Return
    ------
    A disctionary containing AD, DP, cells and variants
    """
    cell_dat = load_VCF(dir_name + "/cellSNP.base.vcf.gz", load_sample=False,
                        biallelic_only=False)
    cell_dat['AD'] = mmread(dir_name + "/cellSNP.tag.AD.mtx").tocsc()
    cell_dat['DP'] = mmread(dir_name + "/cellSNP.tag.DP.mtx").tocsc()
    cell_dat['samples'] = np.genfromtxt(dir_name + "/cellSNP.samples.tsv", dtype=str)
    return cell_dat


def read_vartrix(alt_mtx, ref_mtx, cell_file, vcf_file=None):
    """Read data from VarTrix

    Parameters
    ----------
    alt_mtx:
        sparse matrix file for alternative alleles
    ref_mtx:
        sparse matrix file for reference alleles
    cell_file:
        file for cell barcodes, each per line
    vcf_file:
        the vcf file used for fetch variants in VarTrix
    
    Return
    ------
    A disctionary containing AD, DP, cells and optionally variants
    """
    if vcf_file is not None:
        cell_dat = load_VCF(vcf_file, load_sample=False, biallelic_only=False)
        cell_dat['variants'] = np.array(cell_dat['variants'])
    else:
        cell_dat = {}
    cell_dat['AD'] = mmread(alt_mtx).tocsc()
    cell_dat['DP'] = mmread(ref_mtx).tocsc() + cell_dat['AD']
    cell_dat['samples'] = np.genfromtxt(cell_file, dtype=str)
    return cell_dat


def write_donor_id(out_dir, donor_names, cell_names, n_vars, res_vireo):
    """
    Write the results of donor id into files.
    """
    ID_prob, doublet_prob = res_vireo['ID_prob'], res_vireo['doublet_prob']

    prob_max = np.max(ID_prob, axis=1)
    prob_doublet_out = np.sum(doublet_prob, axis=1)
    donor_singlet = np.array(donor_names, "U100")[np.argmax(ID_prob, axis=1)]

    doublet_names = [",".join(x) for x in permutations(donor_names, 2)]
    donor_doublet = np.array(doublet_names, "U100")[np.argmax(doublet_prob, 
                                                              axis=1)]

    donor_ids = donor_singlet.copy()
    donor_ids[prob_max < 0.9] = "unassigned"
    donor_ids[prob_doublet_out >= 0.9] = "doublet"
    donor_ids[n_vars < 10] = "unassigned"


    ## save log file
    fid = open(out_dir + "/_log.txt", "w")
    fid.writelines("logLik: %.3e\n" %(res_vireo['LB_doublet']))
    fid.writelines("thetas: \n%s\n" %(res_vireo['theta_shapes']))
    fid.close()

    ## save summary file
    fid = open(out_dir + "/summary.tsv", "w")
    fid.writelines("Var1\tFreq\n")
    donor_ids_uniq, donor_ids_count = np.unique(donor_ids, return_counts=True)
    for i in range(len(donor_ids_uniq)):
        fid.writelines("%s\t%d\n" %(donor_ids_uniq[i], donor_ids_count[i]))
    fid.close()
    print("[vireo] final donor size:")
    print("\t".join([str(x) for x in donor_ids_uniq]))
    print("\t".join([str(x) for x in donor_ids_count]))

    ## save donor_ids file
    fid = open(out_dir + "/donor_ids.tsv", "w")
    header = ["cell", "donor_id", "prob_max", "prob_doublet", "n_vars", 
              "best_singlet", "best_doublet"]
    fid.writelines("\t".join(header) + "\n")
    for i in range(len(cell_names)):
        line = [cell_names[i], donor_ids[i], "%.2e" %prob_max[i],
                "%.2e" %prob_doublet_out[i], "%d" %n_vars[i],
                donor_singlet[i], donor_doublet[i]]
        fid.writelines("\t".join(line) + "\n") 
    fid.close()

    ## save singlet probability file
    fid = open(out_dir + "/prob_singlet.tsv", "w")
    fid.writelines("\t".join(["cell"] + donor_names) + "\n")
    for i in range(len(cell_names)):
        line = ["%.2e" %x for x in ID_prob[i, :]]
        fid.writelines("\t".join([cell_names[i]] + line) + "\n") 
    fid.close()

    ## save doublet probability file
    fid = open(out_dir + "/prob_doublet.tsv", "w")
    fid.writelines("\t".join(["cell"] + doublet_names) + "\n")
    for i in range(len(cell_names)):
        line = ["%.2e" %x for x in doublet_prob[i, :]]
        fid.writelines("\t".join([cell_names[i]] + line) + "\n")
    fid.close()

    bashCommand = "gzip -f %s %s" %(out_dir + "/prob_singlet.tsv", 
        out_dir + "/prob_doublet.tsv")
    pro = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    pro.communicate()[0]
